//
// VFO for IC Based Receiver
//
// Based on a port to PIC16F887 of Hans Summers' Si5351a Arduino demo 
// Ported by Andy Fletcher, 2016-07-28 in Edinburgh, Scotland.
// The Si5351A part pulled out into a header for easier insertion into other projects
//
// Original Author: Hans Summers, 2015
// Website: http://www.hanssummers.com
//
// Uses the Si5351a module kit http://www.hanssummers.com/synth
// Please also refer to SiLabs AN619 which describes all the registers to use
//

// TODO
//		Go to sleep and wake on interrupt
//		FIX some tuning turns get ignored

//
// Application Configuration
//
// Minimum frequency in Hz
#define MIN_FREQ 7000000
// Maximum frequency in Hz
#define MAX_FREQ 7020000
// Frequency change per encoder click in Hz
#define FREQ_SLOW_DELTA 100
#define FREQ_FAST_DELTA 1000
// !!!! REMEMBER TO TRIM THE SI_XTAL_FREQ IN SI5351A.H !!!!
// end of application configuration

//
// Configuration bits (PIC16F887) 
// 
#pragma config WDTE = OFF						// disable watchdog
#pragma config PWRTE = OFF						// disable power-up timer
#pragma config CP = OFF							// disable code protection
#pragma config BOREN = ON						// enable brown-out detection	
#pragma config FCMEN = ON						// enable fail-safe clock monitor
#pragma config MCLRE = OFF						// tie ~MCLR to Vdd, free up
#pragma config CPD = OFF							// disable data EEPROM code protection
#pragma config IESO = ON, 						// use internal clock
#pragma config FOSC = INTRC_CLKOUT		// use internal oscillator
#pragma config LVP = OFF							// disable low voltage ICSP

#include <xc.h>
#include <stdint.h>
#include "i2c.h"
#define I2C_SLAVE 0x60                    // Slave device I2C address needed by si5351a.h
#include "si5351a.h"

#define _XTAL_FREQ 8000000
#define DEBUG_LED PORTAbits.RA4
#define POWER_ON_LED PORTDbits.RD4
#define FREQ_LIMIT_LED PORTDbits.RD5

//
// Encoder definitions
//
#define ENCODER         (PORTB & 0x03)
#define ENCODER_BUTTON  PORTBbits.RB2

//
// global variables
//
unsigned char encoderBits, oldEncoderBits;
int encoderCount;
unsigned char encoderButtonState = 0;
unsigned char cw = 0; // clockwise indicator for encoder

//
// ISRs
//
void interrupt handleInterrupt(void) {
   // handle encoder interrupts
   // 
   // When travelling clockwise the encoder bits go     
   // 11 -> 10 -> 00 -> 01 -> 11
   //
   // When travelling anticlockwise the encoder bits go 
   // 11 -> 01 -> 00 -> 10 -> 11
   //
   if (INTCONbits.RBIF) {
        //
        // get latest state of encoder bits
        //
        encoderBits = ENCODER & 0b11;
        // 
        // handle in state machine, ignore 'impossible' changes
        //
        switch (oldEncoderBits) {
          case 0b00: 
            if (encoderBits == 0b01) cw = 1;  // encoder is turning clockwise
            else
            if (encoderBits == 0b10) cw = 0;  // encoder is turning anticlockwise
            break;
          case 0b01:
            if ((encoderBits == 0b11) && (cw == 1)) { 
              // encoder has turned a notch clockwise ??
              cw = 0;
              encoderCount++;
            }
            else
            if (encoderBits == 0b00) cw = 0;  // encoder is turning anticlockwise
            break;
          case 0b10:
            if (encoderBits == 0b00) cw = 1;  // encoder is turning clockwise
            else
            if ((encoderBits == 0b11) && (cw == 0)) { 
              // encoder has turned a notch anticlockwise
              cw = 1;
              encoderCount--;
            }
            break;
          case 0b11:
            if (encoderBits == 0b10) cw = 1;  // encoder is turning clockwise
            else
            if (encoderBits == 0b01) cw = 0;  // encoder is turning anticlockwise
            break;
        }
        oldEncoderBits = encoderBits; // save the state of the encoder

        INTCONbits.RBIF = 0;    // clear interrupt flag
   }
}

//
// production functions
//

void setClock(void) {
  // set PIC internal oscillator to 8MHz
  OSCCONbits.IRCF0 = OSCCONbits.IRCF1 = OSCCONbits.IRCF2 = 1;

  // and tune to around 2MHz
  OSCTUNEbits.TUN4 = 0;
  OSCTUNEbits.TUN3 = 0;
  OSCTUNEbits.TUN2 = 0;
  OSCTUNEbits.TUN1 = 0;
  OSCTUNEbits.TUN0 = 0; 
}

// execution
void main(void)
{
uint32_t frequency, delta;

  // disable all interrupts
  INTCON = 0;
  
  // configure i/o ports
  ANSEL = ANSELH = 0;            // configure all as digital unless needed otherwise
  
  CM1CON0 = CM2CON0 = 0;         // disable comparators
  
  // TODO -- tidy up the PORTs
  TRISA = 0xFF;		    // all of PORTA are inputs
  TRISAbits.TRISA4 = 0;    // except debug led
  
  TRISB = 0xFF;         // all of PORTB are inputs
  
  TRISC = 0xFF;         // all of PORTC are inputs
#ifdef AJFDIAG
  TRISCbits.TRISC6 = 0;    // except EUSART TX for possible debug
#endif
  
  TRISD = 0x00;         // all of PORTD are outputs

  TRISE = 0xFF;  		// all of PORTE are inputs

  setClock();           // set PIC clock to 8MHz
  
  //
  // initialise the encoder
  //
  encoderCount = 0;
  oldEncoderBits = ENCODER;
  IOCB = 0;
  IOCBbits.IOCB0 = IOCBbits.IOCB1 = 1;
  INTCONbits.RBIE = INTCONbits.GIE = 1;		// allow encoder interrupts 
  OPTION_REG = 0b01011110; 					// MSB bit NOT_RBPU cleared to
                                            // enable pull-up on all PORTB pins
  WPUB = 0b00000111;						// weak pull-ups on RB0-2
  
  //
  // initialise the clock generator
  //
  i2c_Init();
  si5351aOutputOff(SI_CLK1_CONTROL);
  si5351aOutputOff(SI_CLK2_CONTROL);

  //
  // initialise LEDs
  //
  POWER_ON_LED = 1;
  FREQ_LIMIT_LED = 0;
  DEBUG_LED = 0;
  __delay_ms(100);

  // set the initial frequency to halfway between the frequency limits
  frequency = MIN_FREQ + ((MAX_FREQ - MIN_FREQ) /2);
  si5351aSetFrequency(frequency);
  
  delta = FREQ_FAST_DELTA; // start with fast tuning
 
  /* loop forever */
  while (1) {
    //
    // handle encoder button
    // 
    encoderButtonState = ((encoderButtonState << 1) | (ENCODER_BUTTON ? 0:1) | 0xe0); // debounce
    if (encoderButtonState == 0xff) {
      encoderButtonState = 0;
      // toggle between fast and slow tuning
      delta = (delta == FREQ_FAST_DELTA ? FREQ_SLOW_DELTA : FREQ_FAST_DELTA); 
    } 
    
    if (encoderCount != 0) {
      // the encoder was turned
      if (cw == 0) {  // TODO -- the code should be == 1 ??
      	frequency += delta;
      	frequency = (frequency > MAX_FREQ ? MAX_FREQ : frequency); // apply limit
      } else {
      	frequency -= delta;
      	frequency = (frequency < MIN_FREQ ? MIN_FREQ : frequency); // apply limit
      }
      if ((frequency == MIN_FREQ) || (frequency == MAX_FREQ)) {
      	FREQ_LIMIT_LED = 1; // indicate limit reached
      } else {
        FREQ_LIMIT_LED = 0;
      }
      si5351aSetFrequency(frequency);
    } 
   
    INTCONbits.RBIE = 0; // reset count with interrupts disabled
    encoderCount = 0;
    INTCONbits.RBIE = 1;
  }
}

